﻿[<AutoOpen>]
module Plot

// Reference WPF dlls
#r "PresentationCore.dll"
#r "PresentationFramework.dll"
#r "WindowsBase.dll"
#r "System.Xaml.dll"
#r "UIAutomationTypes.dll"

open System
open System.Windows
open System.Windows.Media

type Axis (low:double, high:double) =
    member this.Min = low
    member this.Max = high
    member this.Length = high - low

    static member Create (values:seq<float>) =
        let f (l,h) x = min l x, max h x
        Axis(values |> Seq.fold f (0.0,0.0))        

    static member Create (values:seq<float * float>) =
        Axis.Create(values |> Seq.map fst), 
            Axis.Create(values |> Seq.map snd)  

    static member CreateTransform(rect:Rect,xAxis:Axis,yAxis:Axis) =        
        let transform = Matrix.Identity
        transform.Translate(rect.Left, rect.Bottom)        
        transform.ScaleAtPrepend(
            rect.Width / xAxis.Length, 
            -rect.Height / yAxis.Length,
            xAxis.Min, 
            yAxis.Min)
        transform               

type Plot (xAxis,yAxis,render) =
    inherit System.Windows.Controls.ItemsControl ()    

    let CreateText =
        let typeface = new Typeface("Verdana")               
        fun (s, emSize) ->
            new FormattedText(
                s,
                System.Globalization.CultureInfo.InvariantCulture,
                FlowDirection.LeftToRight,
                typeface,
                emSize,
                Brushes.Black)

    override this.OnRender (context) =
        let rect = 
            let width, height =             
                this.ActualWidth-(this.Margin.Left+this.Margin.Right),        
                this.ActualHeight-(this.Margin.Top+this.Margin.Bottom) 
            new Rect(
                this.Margin.Left, 
                this.Margin.Top, 
                max 0.0 width, 
                max 0.0 height)

        let bg = LinearGradientBrush(Colors.White,Colors.LightGray, 90.0)
        context.DrawRectangle(bg, new Pen(Brushes.Black, 1.0), rect)      

        let transform = Axis.CreateTransform (rect, xAxis, yAxis)

        let DrawXLabel x = 
            let label = CreateText(sprintf "%.02f" x, 9.0)
            let p1 = Point(x, yAxis.Min) * transform
            let p2 = Point(x, yAxis.Max) * transform
            context.DrawText(label, 
                new Point(p1.X-(label.Width/2.0), p1.Y+4.0))
            context.DrawText(label,
                new Point(p2.X-(label.Width/2.0), p2.Y-label.Height-4.0))

        let DrawYLabel y =            
            let label = CreateText(sprintf "%.02f" y, 9.0)
            let p1 = Point(xAxis.Min, y) * transform
            let p2 = Point(xAxis.Max, y) * transform
            context.DrawText(label, 
                new Point(p1.X-label.Width-4.0, p2.Y-(label.Height/2.0)))
            context.DrawText(label, 
                new Point(p2.X+4.0, p2.Y-(label.Height/2.0)))        

        DrawXLabel xAxis.Min; DrawXLabel xAxis.Max        
        DrawYLabel yAxis.Min; DrawYLabel yAxis.Max

        render(context,transform)                       

let linePlot (f:(float -> float), xs:seq<float>) =
    let points = xs |> Seq.map (fun x -> x, f x)
    let xAxis, yAxis = Axis.Create points  
    let render (context:DrawingContext,transform:Matrix) =
        let pen = new Pen(Brushes.Red, 1.0)        
        points
        |> Seq.pairwise 
        |> Seq.iter (fun ((x0,y0),(x1,y1)) -> 
            context.DrawLine(pen, 
                new Point(x0,y0) * transform, 
                new Point(x1,y1) * transform)
        ) 
    let plot = new Plot(xAxis,yAxis,render,Margin=Thickness(16.0))
    let win = new Window(Title="Plot",Width=640.0,Height=480.0,Content=plot)
    win.Show() |> ignore

let bubblePlot (points:seq<#Brush * (float * float) * float>) =
    let xAxis, yAxis = 
        points |> Seq.map (fun (_,(x,y),_) -> x,y)
        |> Axis.Create
    let render (context:DrawingContext,transform:Matrix) =
        points
        |> Seq.iter (fun (brush,(x,y),r) ->                              
            let r' = Point(x+r,y+r)*transform - (Point(x-r,y-r)*transform) 
            context.DrawEllipse(brush,null,Point(x,y)*transform,r'.X,r'.Y) 
        )
    let plot = new Plot(xAxis,yAxis,render,Margin=Thickness(16.0))
    let win = new Window(Title="Plot",Width=400.0,Height=400.0,Content=plot)
    win.Show() |> ignore

let red = Brushes.Red
let green = Brushes.Green
let blue = Brushes.Blue